import { RPCs, decodeComponentId, generateComponentId } from '@previewjs/api';
import { createTypeAnalyzer } from '@previewjs/type-analyzer';
import express from 'express';
import fs$1 from 'fs-extra';
import getPort, { portNumbers } from 'get-port';
import { createRequire } from 'module';
import path from 'path';
import { exclusivePromiseRunner } from 'exclusive-promises';
import fs from 'fs';
import url, { URLSearchParams } from 'url';
import { globby } from 'globby';
import { PREVIEW_CONFIG_NAME, readConfig } from '@previewjs/config';
import { createStackedReader, createFileSystemReader } from '@previewjs/vfs';
import assertNever from 'assert-never';
import axios from 'axios';
import { createHttpTerminator } from 'http-terminator';
import { escape } from 'html-escaper';
import fakeExportedTypesPlugin from 'rollup-plugin-friendly-type-imports';
import { loadTsconfig } from 'tsconfig-paths/lib/tsconfig-loader.js';
import * as vite from 'vite';
import { transformWithEsbuild, searchForWorkspaceRoot } from 'vite';
import viteTsconfigPaths from 'vite-tsconfig-paths';
import { Parser } from 'acorn';
import jsx from 'acorn-jsx';

function getCacheDir(rootDirPath) {
  try {
    const __dirname = url.fileURLToPath(new URL(".", import.meta.url));
    const { version } = JSON.parse(
      fs.readFileSync(path.resolve(__dirname, "..", "package.json"), "utf8")
    );
    return path.resolve(
      rootDirPath,
      "node_modules",
      ".previewjs",
      `v${version}`
    );
  } catch (e) {
    throw new Error(`Unable to detect @previewjs/core version.
${e}`);
  }
}

async function findFiles(rootDirPath, pattern) {
  const gitRootPath = await findGitRoot(rootDirPath);
  const relativePath = path.relative(gitRootPath, rootDirPath);
  const relativePrefix = relativePath ? relativePath + path.sep : "";
  const files = await globby(relativePrefix + pattern, {
    gitignore: true,
    ignore: ["**/node_modules/**"],
    cwd: gitRootPath,
    absolute: true,
    followSymbolicLinks: false
  });
  return files.filter(
    (f) => f.startsWith(rootDirPath.replace(/\\/g, "/") + "/")
  );
}
async function findGitRoot(dirPath, fallback = dirPath) {
  try {
    if (await fs$1.exists(path.join(dirPath, ".git"))) {
      return dirPath;
    } else {
      const parentDirPath = path.dirname(dirPath);
      if (!parentDirPath || parentDirPath === dirPath) {
        return fallback;
      }
      return findGitRoot(parentDirPath, fallback);
    }
  } catch {
    return fallback;
  }
}

const FILES_REQUIRING_REDETECTION = /* @__PURE__ */ new Set([
  "jsconfig.json",
  "tsconfig.json",
  "package.json",
  "package-lock.json",
  "pnpm-lock.yaml",
  "yarn.lock"
]);
const oneAtATime = exclusivePromiseRunner();
function detectComponents(workspace, frameworkPlugin, typeAnalyzer, options = {}) {
  return oneAtATime(async () => {
    const detectionStartTimestamp = Date.now();
    const cacheFilePath = path.join(
      getCacheDir(workspace.rootDirPath),
      "components.json"
    );
    const absoluteFilePaths = options.filePaths ? options.filePaths.map(
      (filePath) => path.join(workspace.rootDirPath, filePath)
    ) : await findFiles(
      workspace.rootDirPath,
      "**/*.@(js|jsx|ts|tsx|svelte|vue)"
    );
    const filePathsSet = new Set(
      absoluteFilePaths.map(
        (absoluteFilePath) => path.relative(workspace.rootDirPath, absoluteFilePath).replace(/\\/g, "/")
      )
    );
    let existingCache = fs$1.existsSync(cacheFilePath) ? JSON.parse(
      fs$1.readFileSync(cacheFilePath, "utf8")
    ) : {
      detectionStartTimestamp: 0,
      components: {}
    };
    if (existingCache.detectionStartTimestamp < await detectionMinimalTimestamp(workspace.rootDirPath)) {
      existingCache = {
        detectionStartTimestamp: 0,
        components: {}
      };
    }
    const changedAbsoluteFilePaths = absoluteFilePaths.filter(
      (absoluteFilePath) => {
        const entry = workspace.reader.readSync(absoluteFilePath);
        return entry?.kind === "file" && (options.forceRefresh || entry.lastModifiedMillis() >= existingCache.detectionStartTimestamp);
      }
    );
    const recycledComponents = Object.fromEntries(
      Object.entries(existingCache.components).filter(
        ([filePath]) => filePathsSet.has(filePath)
      )
    );
    const refreshedComponents = await detectComponentsCore(
      workspace,
      frameworkPlugin,
      typeAnalyzer,
      changedAbsoluteFilePaths
    );
    const allComponents = {
      ...recycledComponents,
      ...refreshedComponents
    };
    const components = Object.keys(allComponents).sort().reduce((ordered, filePath) => {
      ordered[filePath] = allComponents[filePath];
      return ordered;
    }, {});
    if (!options.filePaths) {
      await fs$1.mkdirp(path.dirname(cacheFilePath));
      const updatedCache = {
        detectionStartTimestamp,
        components
      };
      await fs$1.writeFile(cacheFilePath, JSON.stringify(updatedCache));
    }
    return { components };
  });
}
async function detectComponentsCore(workspace, frameworkPlugin, typeAnalyzer, changedAbsoluteFilePaths) {
  const components = {};
  if (changedAbsoluteFilePaths.length === 0) {
    return components;
  }
  const found = await frameworkPlugin.detectComponents(
    workspace.reader,
    typeAnalyzer,
    changedAbsoluteFilePaths
  );
  for (const component of found) {
    const filePath = path.relative(workspace.rootDirPath, component.absoluteFilePath).replace(/\\/g, "/");
    const fileComponents = components[filePath] || (components[filePath] = []);
    fileComponents.push(
      detectedComponentToApiComponent(workspace.rootDirPath, component)
    );
  }
  return components;
}
function detectedComponentToApiComponent(rootDirPath, component) {
  const [start, end] = component.offsets[0];
  return {
    name: component.name,
    start,
    end,
    info: component.info.kind === "component" ? {
      kind: "component",
      exported: component.info.exported
    } : {
      kind: "story",
      args: component.info.args,
      associatedComponent: {
        filePath: path.relative(
          rootDirPath,
          component.info.associatedComponent.absoluteFilePath
        ),
        name: component.info.associatedComponent.name
      }
    }
  };
}
async function detectionMinimalTimestamp(rootDirPath) {
  const nodeModulesPath = path.join(rootDirPath, "node_modules");
  let lastModifiedMillis = 0;
  if (await fs$1.pathExists(nodeModulesPath)) {
    for (const subdirectory of await fs$1.readdir(nodeModulesPath)) {
      const subdirectoryPath = path.join(nodeModulesPath, subdirectory);
      const stat = await fs$1.lstat(subdirectoryPath);
      if (stat.isDirectory() || stat.isSymbolicLink()) {
        lastModifiedMillis = Math.max(lastModifiedMillis, stat.mtimeMs);
      }
    }
  }
  for (const f of FILES_REQUIRING_REDETECTION) {
    const filePath = path.join(rootDirPath, f);
    if (await fs$1.pathExists(filePath)) {
      const stat = await fs$1.stat(filePath);
      lastModifiedMillis = Math.max(lastModifiedMillis, stat.mtimeMs);
    }
  }
  return lastModifiedMillis;
}

class Server {
  constructor(options) {
    this.serverTerminator = null;
    this.server = null;
    const app = express();
    app.use((_req, res, next) => {
      res.setHeader("Cache-Control", "max-age=0, must-revalidate");
      next();
    });
    for (const middleware of options.middlewares) {
      app.use(middleware);
    }
    this.app = app;
  }
  async start(port) {
    return new Promise((resolve) => {
      const server = this.server = this.app.listen(port, () => {
        console.log(`Preview.js Server running at http://localhost:${port}.`);
        resolve(server);
      });
      this.serverTerminator = createHttpTerminator({
        server: this.server,
        gracefulTerminationTimeout: 0
      });
    });
  }
  async stop() {
    if (this.serverTerminator) {
      await this.serverTerminator.terminate();
      this.serverTerminator = null;
    }
    if (this.server) {
      const server = this.server;
      await new Promise((resolve) => server.close(resolve));
      this.server = null;
    }
    console.log(`Preview.js Server stopped.`);
  }
}

const COMPONENT_LOADER_MODULE = "/@component-loader.js";
function componentLoaderPlugin(options) {
  return {
    name: "previewjs:component-loader",
    resolveId: async function(id) {
      if (id.startsWith(COMPONENT_LOADER_MODULE)) {
        return id;
      }
      return null;
    },
    load: async function(id) {
      if (!id.startsWith(COMPONENT_LOADER_MODULE)) {
        return null;
      }
      const params = new URLSearchParams(id.split("?")[1] || "");
      return generateComponentLoaderModule(params, options);
    }
  };
}
function generateComponentLoaderModule(urlParams, {
  rootDirPath,
  detectedGlobalCssFilePaths,
  config: { wrapper }
}) {
  const filePath = urlParams.get("p");
  const componentName = urlParams.get("c");
  if (filePath === null || componentName === null) {
    throw new Error(`Invalid use of ${COMPONENT_LOADER_MODULE} module`);
  }
  const componentModuleId = `/${filePath.replace(/\\/g, "/")}`;
  return `import { updateComponent } from '/__previewjs_internal__/update-component';
import { load } from '/__previewjs_internal__/renderer/index';

let componentLoadId;
let getLatestComponentLoadId;
let refreshId = 0;

export function init(options) {
  ({
    componentLoadId,
    getLatestComponentLoadId,
  } = options);
}

export async function refresh() {
  const renderId = componentLoadId + "-" + (++refreshId);
  const shouldAbortRender = () => renderId !== (getLatestComponentLoadId() + "-" + refreshId);
  let loadingError = null;
  ${wrapper && fs$1.pathExistsSync(path.join(rootDirPath, wrapper.path)) ? `
  let wrapperModulePromise;
  if (import.meta.hot.data.preloadedWrapperModule) {
    wrapperModulePromise = Promise.resolve(import.meta.hot.data.preloadedWrapperModule);
  } else {
    wrapperModulePromise = import("/${wrapper.path}");
  }
  const wrapperModule = await wrapperModulePromise.catch(e => {
    console.error(e);
    loadingError = e.stack || e.message || null;
    return null;
  });
  ` : `
  ${detectedGlobalCssFilePaths.map(
    (filePath2) => `import("/${filePath2.replace(/\\/g, "/")}").catch(console.warn);`
  ).join("\n")}
  const wrapperModule = null;
  `}
  let componentModulePromise;
  if (import.meta.hot.data.preloadedComponentModule) {
    componentModulePromise = Promise.resolve(import.meta.hot.data.preloadedComponentModule);
  } else {
    componentModulePromise = import("${componentModuleId}");
  }
  const componentModule = await componentModulePromise.catch(e => {
    console.error(e);
    loadingError = e.stack || e.message || null;
    return null;
  });
  if (shouldAbortRender()) {
    return;
  }
  await updateComponent({
    wrapperModule,
    wrapperName: ${JSON.stringify(wrapper?.componentName || null)},
    componentModule,
    componentFilePath: ${JSON.stringify(filePath)},
    componentName: ${JSON.stringify(componentName)},
    renderId,
    shouldAbortRender,
    loadingError,
    load,
  })
}

import.meta.hot.accept();

${wrapper ? `
import.meta.hot.accept(["${wrapper.path}"], ([wrapperModule]) => {
  if (wrapperModule) {
    import.meta.hot.data.preloadedWrapperModule = wrapperModule;
    refresh();
  }
});
` : ``}

import.meta.hot.accept(["${componentModuleId}"], ([componentModule]) => {
  if (componentModule) {
    import.meta.hot.data.preloadedComponentModule = componentModule;
    refresh();
  }
});
`;
}

const extensions = /* @__PURE__ */ new Set([".js", ".jsx", ".ts", ".tsx"]);
const importRegexp = /import\s+([\S]+)\s+from\s+('|")([\S]+)\.(css|less|sass|scss|styl|stylus|postcss)(\?[\S]*)?&?('|")/g;
function cssModulesWithoutSuffixPlugin() {
  return {
    name: "previewjs:css-modules-without-suffix",
    async transform(code, id) {
      id = id.split("?")[0];
      if (!extensions.has(path.extname(id))) {
        return;
      }
      return code.replace(
        importRegexp,
        'import $1 from "$3.$4?$5&.module.css"'
      );
    }
  };
}

const jsExtensions$1 = /* @__PURE__ */ new Set([".js", ".jsx", ".ts", ".tsx"]);
function exportToplevelPlugin() {
  return {
    name: "previewjs:export-toplevel",
    enforce: "post",
    transform: async function(code, id) {
      if (id.indexOf(`/node_modules/`) !== -1) {
        return null;
      }
      id = id.split("?", 2)[0];
      const extension = path.extname(id);
      if (!jsExtensions$1.has(extension)) {
        return null;
      }
      try {
        const topLevelEntityNames = findTopLevelEntityNames(code);
        return `${code};
          export {
          ${topLevelEntityNames.map((c) => `${c} as __previewjs__${c},`).join("")}
        }`;
      } catch (e) {
        throw new Error(`Unable to parse ${id}: ${e}`);
      }
    }
  };
}
function findTopLevelEntityNames(source) {
  const parsed = Parser.extend(jsx()).parse(source, {
    ecmaVersion: "latest",
    sourceType: "module"
  });
  const topLevelEntityNames = [];
  for (const statement of parsed.body || []) {
    if (statement.type === "VariableDeclaration") {
      for (const declaration of statement.declarations) {
        if (declaration.type === "VariableDeclarator") {
          addIfIdentifier(topLevelEntityNames, declaration.id);
        }
      }
    }
    if (statement.type === "FunctionDeclaration") {
      addIfIdentifier(topLevelEntityNames, statement.id);
    }
    if (statement.type === "ClassDeclaration") {
      addIfIdentifier(topLevelEntityNames, statement.id);
    }
    if (statement.type === "ExportDefaultDeclaration") {
      if (statement.declaration) {
        addIfIdentifier(topLevelEntityNames, statement.declaration);
      }
    }
    if (statement.type === "ExportNamedDeclaration") {
      if (statement.declaration?.id) {
        addIfIdentifier(topLevelEntityNames, statement.declaration.id);
      }
      if (statement.declaration?.declarations) {
        for (const declarator of statement.declaration.declarations) {
          addIfIdentifier(topLevelEntityNames, declarator.id);
        }
      }
    }
  }
  return [...new Set(topLevelEntityNames)];
}
function addIfIdentifier(array, id) {
  if (id && id.type === "Identifier") {
    const name = id.name;
    if (name.endsWith("_default")) {
      return;
    }
    array.push(name);
  }
}

function localEval() {
  return {
    name: "previewjs:local-eval",
    enforce: "post",
    transform: (code, id) => {
      if (id.indexOf(`/node_modules/`) !== -1) {
        return null;
      }
      if (id.indexOf("?") !== -1) {
        return null;
      }
      return code + // Note: This is an uppercase function so that it "looks like" a React component.
      // This needed to prevent breaking React Refresh, which expects all exports to
      // be React components.
      `
export const PreviewJsEvaluateLocally = async (autogenCallbackPropsSource, propsAssignmentSource, __jsxFactory__) => {
  let autogenCallbackProps = {};
  eval(autogenCallbackPropsSource);
  let properties = {};
  eval(propsAssignmentSource);
  return { autogenCallbackProps, properties };
}
`;
    }
  };
}

const VIRTUAL_PREFIX = `/@previewjs-virtual:`;
const jsExtensions = /* @__PURE__ */ new Set([".js", ".jsx", ".ts", ".tsx"]);
function virtualPlugin(options) {
  const { reader, rootDirPath } = options;
  return {
    name: "previewjs:virtual-fs",
    resolveId: async function(id, importer) {
      const virtualImporter = importer?.startsWith(VIRTUAL_PREFIX) || false;
      if (id.indexOf(`/node_modules/`) !== -1) {
        return null;
      }
      if (id.startsWith(VIRTUAL_PREFIX)) {
        id = id.slice(VIRTUAL_PREFIX.length);
      }
      id = id.split("?", 2)[0];
      const extension = path.extname(id);
      let absoluteId;
      if (path.isAbsolute(id)) {
        absoluteId = id;
      } else {
        if (!importer) {
          return null;
        }
        if (virtualImporter) {
          importer = importer.slice(VIRTUAL_PREFIX.length);
        }
        if (extension && !jsExtensions.has(extension) && extension !== ".svg") {
          return path.join(path.dirname(importer), id);
        }
        absoluteId = path.join(path.dirname(importer), id);
      }
      const resolved = await resolveAbsoluteModuleId(absoluteId);
      if (resolved) {
        const [absoluteFilePath] = resolved;
        if (await fs$1.pathExists(absoluteFilePath) && !virtualImporter) {
          return null;
        }
        if (!absoluteId.endsWith(extension)) {
          absoluteId += extension;
        }
        const resolvedId = VIRTUAL_PREFIX + absoluteFilePath.replace(/\\/g, "/");
        return resolvedId;
      }
      return null;
    },
    load: async function(id) {
      if (id.indexOf(`/node_modules/`) !== -1) {
        return null;
      }
      if (id.startsWith(VIRTUAL_PREFIX)) {
        id = id.slice(VIRTUAL_PREFIX.length);
      }
      const resolved = await resolveAbsoluteModuleId(id);
      if (!resolved) {
        return null;
      }
      const [absoluteFilePath, entry] = resolved;
      let isAllowed = false;
      for (const allowedAbsolutePath of options.allowedAbsolutePaths) {
        const relativePath = path.relative(
          allowedAbsolutePath,
          absoluteFilePath
        );
        if (relativePath && !relativePath.startsWith("..") && !path.isAbsolute(relativePath)) {
          isAllowed = true;
          break;
        }
      }
      if (!isAllowed) {
        console.error(
          `Attempted access to ${absoluteFilePath} which is outside of allowed directories. See https://previewjs.com/docs/config and https://vitejs.dev/config/server-options.html#server-fs-allow for more information.`
        );
        return null;
      }
      if (entry.kind !== "file") {
        console.error(`Unable to read file from ${absoluteFilePath}`);
        return null;
      }
      const source = await entry.read();
      const fileExtension = path.extname(absoluteFilePath);
      if (!jsExtensions.has(fileExtension)) {
        return source;
      }
      const moduleExtension = path.extname(id);
      return {
        code: (
          // We run an esbuild transform for .js or no extension
          // because it may include JSX. Otherwise, let plugins
          // decide whether to use esbuild or not.
          !id.includes("__previewjs_internal__") && (moduleExtension === "" || moduleExtension === ".js") ? (await transformWithEsbuild(source, absoluteFilePath, {
            loader: "tsx",
            format: "esm",
            target: "es2020",
            sourcefile: path.relative(rootDirPath, absoluteFilePath),
            ...options.esbuildOptions
          })).code : source
        )
      };
    },
    transform: (code, id) => {
      if (!id.startsWith(VIRTUAL_PREFIX)) {
        return null;
      }
      return {
        code,
        map: {
          mappings: "",
          names: [],
          sources: [],
          version: 3
        }
      };
    },
    handleHotUpdate: async function(context) {
      const moduleGraph = options.moduleGraph();
      if (!moduleGraph) {
        return;
      }
      const filePath = path.relative(rootDirPath, context.file);
      const absoluteFilePath = path.join(rootDirPath, filePath);
      const entry = await reader.read(absoluteFilePath);
      if (!entry || entry.kind !== "file") {
        return;
      }
      const virtualModuleId = VIRTUAL_PREFIX + absoluteFilePath.replace(/\\/g, "/");
      const node = moduleGraph.getModuleById(virtualModuleId);
      return node && [node];
    }
  };
  async function resolveAbsoluteModuleId(id) {
    return await resolveBaseFilePath(id) || await resolveBaseFilePath(path.join(rootDirPath, id));
  }
  async function resolveBaseFilePath(baseFilePath) {
    for (const suffix of [
      "",
      ...jsExtensions,
      ...[...jsExtensions].map((extension) => `${path.sep}index${extension}`)
    ]) {
      const absoluteFilePath = `${baseFilePath}${suffix}`;
      const entry = await reader.read(absoluteFilePath);
      if (entry?.kind === "file") {
        return [absoluteFilePath, entry];
      }
    }
    return null;
  }
}

class ViteManager {
  constructor(options) {
    this.options = options;
    this.lastPingTimestamp = 0;
    const router = express.Router();
    router.get("/preview/", async (req, res) => {
      try {
        const template = await fs$1.readFile(
          this.options.shadowHtmlFilePath,
          "utf-8"
        );
        await this.viteStartupPromise;
        if (!this.viteServer) {
          res.status(404).end(`Uh-Oh! Vite server is not running.`);
          return;
        }
        res.status(200).set({ "Content-Type": "text/html" }).end(
          await this.viteServer.transformIndexHtml(req.originalUrl, template)
        );
      } catch (e) {
        res.status(500).set({ "Content-Type": "text/html" }).end(
          `<html>
              <head>
                <style>
                  body {
                    background: #FCA5A5
                  }
                  pre {
                    font-family: source-code-pro, Menlo, Monaco, Consolas, 'Courier New',
                    monospace;
                    font-size: 12px;
                    line-height: 1.5em;
                    color: #7F1D1D;
                  }
                </style>
              </head>
              <body>
                <pre>${escape(`${e}` || "An unknown error has occurred")}</pre>
              </body>
            </html>`
        );
      }
    });
    router.use("/ping", async (req, res) => {
      this.lastPingTimestamp = Date.now();
      res.json(
        JSON.stringify({
          pong: "match!"
        })
      );
    });
    router.use(async (req, res, next) => {
      const waitSeconds = 60;
      const waitUntil = Date.now() + waitSeconds * 1e3;
      while (!this.viteServer) {
        await new Promise((resolve) => setTimeout(resolve, 50));
        if (Date.now() > waitUntil) {
          throw new Error(
            `Vite server is not running after ${waitSeconds} seconds.`
          );
        }
      }
      this.viteServer.middlewares(req, res, next);
    });
    this.middleware = router;
  }
  async start(server, port) {
    const typeScriptConfigAbsoluteFilePaths = await findFiles(
      this.options.rootDirPath,
      "{js,ts}config.json"
    );
    const typeScriptConfigFilePaths = typeScriptConfigAbsoluteFilePaths.map(
      (p) => path.relative(this.options.rootDirPath, p)
    );
    const validTypeScriptFilePaths = [];
    for (const configFilePath of typeScriptConfigFilePaths) {
      try {
        loadTsconfig(configFilePath);
        validTypeScriptFilePaths.push(configFilePath);
      } catch (e) {
        console.warn(
          `Encountered an invalid config file, ignoring: ${configFilePath}`
        );
      }
    }
    const tsInferredAlias = [];
    const config = loadTsconfig(
      typeScriptConfigFilePaths.includes("tsconfig.json") ? "tsconfig.json" : "jsconfig.json"
    );
    if (config?.compilerOptions?.baseUrl && config?.compilerOptions?.paths) {
      const { baseUrl, paths } = config.compilerOptions;
      for (const [match, mapping] of Object.entries(paths)) {
        const firstMapping = mapping[0];
        if (!firstMapping) {
          continue;
        }
        const matchNoWildcard = match.endsWith("/*") ? match.slice(0, match.length - 2) : match;
        const firstMappingNoWildcard = firstMapping.endsWith("/*") ? firstMapping.slice(0, firstMapping.length - 2) : firstMapping;
        tsInferredAlias.push({
          find: matchNoWildcard,
          replacement: path.join(
            this.options.rootDirPath,
            baseUrl,
            firstMappingNoWildcard
          )
        });
      }
    }
    const existingViteConfig = await vite.loadConfigFromFile(
      {
        command: "serve",
        mode: "development"
      },
      void 0,
      this.options.rootDirPath
    );
    const defaultLogger = vite.createLogger(this.options.logLevel);
    const frameworkPluginViteConfig = this.options.frameworkPlugin.viteConfig(
      await flattenPlugins([
        ...existingViteConfig?.config.plugins || [],
        ...this.options.config.vite?.plugins || []
      ])
    );
    const vitePlugins = [
      viteTsconfigPaths({
        root: this.options.rootDirPath,
        projects: validTypeScriptFilePaths
      }),
      virtualPlugin({
        reader: this.options.reader,
        rootDirPath: this.options.rootDirPath,
        allowedAbsolutePaths: this.options.config.vite?.server?.fs?.allow || [
          searchForWorkspaceRoot(this.options.rootDirPath)
        ],
        moduleGraph: () => this.viteServer?.moduleGraph || null,
        esbuildOptions: frameworkPluginViteConfig.esbuild || {}
      }),
      localEval(),
      exportToplevelPlugin(),
      fakeExportedTypesPlugin({
        readFile: (absoluteFilePath) => this.options.reader.read(absoluteFilePath).then((entry) => {
          if (entry?.kind !== "file") {
            return null;
          }
          return entry.read();
        })
      }),
      cssModulesWithoutSuffixPlugin(),
      componentLoaderPlugin(this.options),
      frameworkPluginViteConfig.plugins
    ];
    const plugins = await Promise.all(
      vitePlugins.flat().map(async (pluginOrPromise) => {
        const plugin = await pluginOrPromise;
        if (!plugin || Array.isArray(plugin) || !plugin.handleHotUpdate) {
          return plugin;
        }
        const handleHotUpdate = typeof plugin.handleHotUpdate === "function" ? plugin.handleHotUpdate : plugin.handleHotUpdate.handler;
        return {
          ...plugin,
          handleHotUpdate: async (ctx) => {
            return handleHotUpdate({
              ...ctx,
              read: async () => {
                const entry = await this.options.reader.read(ctx.file);
                if (entry?.kind !== "file") {
                  return ctx.read();
                }
                return entry.read();
              }
            });
          }
        };
      })
    );
    const viteServerPromise = vite.createServer({
      ...frameworkPluginViteConfig,
      ...existingViteConfig?.config,
      ...this.options.config.vite,
      configFile: false,
      root: this.options.rootDirPath,
      base: "/preview/",
      server: {
        middlewareMode: true,
        hmr: {
          overlay: false,
          server,
          clientPort: port,
          ...typeof this.options.config.vite?.server?.hmr === "object" ? this.options.config.vite?.server?.hmr : {}
        },
        fs: {
          strict: false,
          ...this.options.config.vite?.server?.fs || {}
        },
        ...this.options.config.vite?.server
      },
      customLogger: {
        info: defaultLogger.info,
        warn: defaultLogger.warn,
        error: (msg, options) => {
          if (!msg.startsWith("\x1B[31mInternal server error")) {
            this.viteServer?.ws.send({
              type: "error",
              err: {
                message: msg,
                stack: ""
              }
            });
          }
          defaultLogger.error(msg, options);
        },
        warnOnce: defaultLogger.warnOnce,
        clearScreen: () => {
        },
        hasWarned: defaultLogger.hasWarned,
        hasErrorLogged: defaultLogger.hasErrorLogged
      },
      clearScreen: false,
      cacheDir: this.options.config.vite?.cacheDir || existingViteConfig?.config.cacheDir || this.options.cacheDir,
      publicDir: this.options.config.vite?.publicDir || existingViteConfig?.config.publicDir || frameworkPluginViteConfig.publicDir || this.options.config.publicDir,
      plugins,
      define: {
        ...existingViteConfig?.config.define,
        __filename: void 0,
        __dirname: void 0,
        ...frameworkPluginViteConfig.define,
        ...existingViteConfig?.config.define,
        ...this.options.config.vite?.define
      },
      resolve: {
        ...existingViteConfig?.config.resolve,
        ...this.options.config.vite?.resolve,
        alias: [
          // First defined rules are applied first, therefore highest priority should come first.
          ...viteAliasToRollupAliasEntries(
            this.options.config.vite?.resolve?.alias
          ),
          ...viteAliasToRollupAliasEntries(this.options.config.alias),
          ...viteAliasToRollupAliasEntries(
            existingViteConfig?.config.resolve?.alias
          ),
          ...tsInferredAlias,
          {
            find: "~",
            replacement: ""
          },
          {
            find: "@",
            replacement: ""
          },
          ...viteAliasToRollupAliasEntries(
            frameworkPluginViteConfig.resolve?.alias
          )
        ]
      }
    });
    this.viteStartupPromise = new Promise((resolve) => {
      viteServerPromise.finally(() => {
        resolve();
        delete this.viteStartupPromise;
      });
    });
    this.viteServer = await viteServerPromise;
    this.viteServer.watcher.addListener("change", (path2) => {
      this.viteServer?.ws.send({
        type: "custom",
        event: "previewjs-file-changed",
        data: {
          path: path2
        }
      });
    });
  }
  async stop() {
    if (this.viteServer) {
      await this.viteServer.close();
      delete this.viteServer;
    }
  }
  getConfig() {
    return this.options.config;
  }
  getLastPingTimestamp() {
    return this.lastPingTimestamp;
  }
  triggerReload(absoluteFilePath) {
    if (!this.viteServer) {
      return;
    }
    for (const onChange of this.viteServer.watcher.listeners("change")) {
      onChange(absoluteFilePath);
    }
  }
  triggerFullReload() {
    this.viteServer?.moduleGraph.invalidateAll();
    this.viteServer?.ws.send({
      type: "full-reload"
    });
  }
}
function viteAliasToRollupAliasEntries(alias) {
  if (!alias) {
    return [];
  }
  if (Array.isArray(alias)) {
    return alias;
  } else {
    return Object.entries(alias).map(([find, replacement]) => ({
      find,
      replacement
    }));
  }
}
async function flattenPlugins(pluginOptions) {
  const plugins = [];
  for (const pluginOption of await Promise.all(pluginOptions)) {
    if (!pluginOption) {
      continue;
    }
    if (Array.isArray(pluginOption)) {
      plugins.push(...await flattenPlugins(pluginOption));
    } else {
      plugins.push(pluginOption);
    }
  }
  return plugins;
}

const POSTCSS_CONFIG_FILE = [
  ".postcssrc",
  ".postcssrc.json",
  ".postcssrc.yml",
  ".postcssrc.js",
  ".postcssrc.mjs",
  ".postcssrc.cjs",
  ".postcssrc.ts",
  "postcss.config.js",
  "postcss.config.mjs",
  "postcss.config.cjs",
  "postcss.config.ts"
];
const GLOBAL_CSS_FILE_NAMES_WITHOUT_EXT = [
  "index",
  "global",
  "globals",
  "style",
  "styles",
  "app"
];
const GLOBAL_CSS_EXTS = ["css", "sass", "scss", "less", "styl", "stylus"];
const GLOBAL_CSS_FILE = GLOBAL_CSS_FILE_NAMES_WITHOUT_EXT.flatMap(
  (fileName) => GLOBAL_CSS_EXTS.map((ext) => `${fileName}.${ext}`)
);
const FILES_REQUIRING_RESTART = /* @__PURE__ */ new Set([
  PREVIEW_CONFIG_NAME,
  ...FILES_REQUIRING_REDETECTION,
  ...POSTCSS_CONFIG_FILE,
  ...GLOBAL_CSS_FILE,
  "vite.config.js",
  "vite.config.ts"
]);
const SHUTDOWN_CHECK_INTERVAL = 3e3;
const SHUTDOWN_AFTER_INACTIVITY = 1e4;
class Previewer {
  constructor(options) {
    this.options = options;
    this.appServer = null;
    this.viteManager = null;
    this.status = { kind: "stopped" };
    this.shutdownCheckInterval = null;
    this.disposeObserver = null;
    this.config = null;
    this.onFileChangeListener = {
      onChange: (absoluteFilePath, info) => {
        absoluteFilePath = path.resolve(absoluteFilePath);
        if (!info.virtual && this.config?.wrapper && absoluteFilePath === path.resolve(this.options.rootDirPath, this.config.wrapper.path)) {
          this.viteManager?.triggerFullReload();
        }
        if (!info.virtual && FILES_REQUIRING_RESTART.has(path.basename(absoluteFilePath))) {
          if (this.status.kind === "starting" || this.status.kind === "started") {
            const port = this.status.port;
            console.log("New dependencies were detected. Restarting...");
            this.stop({
              restarting: true
            }).then(async () => {
              await this.start(async () => port, { restarting: true });
            }).catch(console.error);
          }
          return;
        }
        if (info.virtual) {
          this.viteManager?.triggerReload(absoluteFilePath);
          this.viteManager?.triggerReload(absoluteFilePath + ".ts");
        } else if (this.options.onFileChanged) {
          this.options.onFileChanged(absoluteFilePath);
        }
      }
    };
    this.transformingReader = createStackedReader([
      options.reader,
      createFileSystemReader({
        mapping: {
          from: options.frameworkPlugin.previewDirPath,
          to: path.join(
            options.rootDirPath,
            "__previewjs_internal__",
            "renderer"
          )
        },
        watch: false
      }),
      createFileSystemReader({
        mapping: {
          from: options.previewDirPath,
          to: options.rootDirPath
        },
        watch: false
      })
    ]);
  }
  async start(allocatePort, options = {}) {
    let port;
    const statusBeforeStart = this.status;
    switch (statusBeforeStart.kind) {
      case "starting":
        port = statusBeforeStart.port;
        try {
          await statusBeforeStart.promise;
        } catch (e) {
          console.error(e);
          this.status = {
            kind: "stopped"
          };
          await this.startFromStopped(port);
        }
        break;
      case "started":
        if (this.shutdownCheckInterval) {
          clearInterval(this.shutdownCheckInterval);
          this.shutdownCheckInterval = null;
        }
        port = statusBeforeStart.port;
        break;
      case "stopping":
        port = statusBeforeStart.port;
        try {
          await statusBeforeStart.promise;
        } catch (e) {
          console.error(e);
          this.status = {
            kind: "stopped"
          };
        }
        await this.start(async () => port, options);
        break;
      case "stopped":
        port = await allocatePort();
        await this.startFromStopped(port, options);
        break;
      default:
        throw assertNever(statusBeforeStart);
    }
    return port;
  }
  async startFromStopped(port, { restarting } = {}) {
    this.status = {
      kind: "starting",
      port,
      promise: (async () => {
        process.chdir(this.options.rootDirPath);
        const config = await readConfig(this.options.rootDirPath);
        this.config = {
          ...config,
          wrapper: config.wrapper || {
            path: this.options.frameworkPlugin.defaultWrapperPath
          }
        };
        const globalCssAbsoluteFilePaths = await findFiles(
          this.options.rootDirPath,
          `**/@(${GLOBAL_CSS_FILE_NAMES_WITHOUT_EXT.join(
            "|"
          )}).@(${GLOBAL_CSS_EXTS.join("|")})`
        );
        this.appServer = new Server({
          middlewares: [
            ...this.options.middlewares || [],
            (req, res, next) => {
              this.viteManager?.middleware(req, res, next);
            }
          ]
        });
        if (!restarting) {
          if (this.transformingReader.observe) {
            this.disposeObserver = await this.transformingReader.observe(
              this.options.rootDirPath
            );
          }
          this.transformingReader.listeners.add(this.onFileChangeListener);
        }
        this.viteManager = new ViteManager({
          rootDirPath: this.options.rootDirPath,
          shadowHtmlFilePath: path.join(
            this.options.previewDirPath,
            "index.html"
          ),
          detectedGlobalCssFilePaths: globalCssAbsoluteFilePaths.map(
            (absoluteFilePath) => path.relative(this.options.rootDirPath, absoluteFilePath)
          ),
          reader: this.transformingReader,
          cacheDir: path.join(getCacheDir(this.options.rootDirPath), "vite"),
          config: this.config,
          logLevel: this.options.logLevel,
          frameworkPlugin: this.options.frameworkPlugin
        });
        const server = await this.appServer.start(port);
        await this.viteManager.start(server, port);
        this.status = {
          kind: "started",
          port
        };
      })()
    };
    await this.status.promise;
    for (let i = 0; ; i++) {
      try {
        await axios.get(`http://localhost:${port}`);
        break;
      } catch (e) {
        if (i === 10) {
          throw e;
        } else {
          await new Promise((resolve) => setTimeout(resolve, 50));
        }
      }
    }
  }
  async stop(options = {}) {
    if (this.status.kind === "starting") {
      try {
        await this.status.promise;
      } catch {
        this.status = {
          kind: "stopped"
        };
      }
    }
    if (this.status.kind !== "started") {
      return;
    }
    if (options.onceUnused) {
      if (!await this.shutdownCheck()) {
        this.shutdownCheckInterval = setInterval(() => {
          this.shutdownCheck().catch(console.error);
        }, SHUTDOWN_CHECK_INTERVAL);
      }
    } else {
      await this.stopNow(options);
    }
  }
  async shutdownCheck() {
    const lastPingTimestamp = this.viteManager?.getLastPingTimestamp() || 0;
    if (lastPingTimestamp + SHUTDOWN_AFTER_INACTIVITY > Date.now()) {
      return false;
    }
    if (this.shutdownCheckInterval) {
      clearInterval(this.shutdownCheckInterval);
      this.shutdownCheckInterval = null;
    }
    await this.stopNow();
    return true;
  }
  async stopNow({
    restarting
  } = {}) {
    if (this.status.kind === "starting") {
      await this.status.promise;
    }
    if (this.status.kind !== "started") {
      return;
    }
    this.status = {
      kind: "stopping",
      port: this.status.port,
      promise: (async () => {
        if (!restarting) {
          this.transformingReader.listeners.remove(this.onFileChangeListener);
          if (this.disposeObserver) {
            await this.disposeObserver();
            this.disposeObserver = null;
          }
        }
        if (this.viteManager) {
          await this.viteManager.stop();
          this.viteManager = null;
        }
        if (this.appServer) {
          await this.appServer.stop();
          this.appServer = null;
        }
        this.status = {
          kind: "stopped"
        };
      })()
    };
    await this.status.promise;
  }
}

class ApiRouter {
  constructor() {
    this.handlers = /* @__PURE__ */ new Map();
    this.registerRPC = (endpoint, handler) => {
      if (this.handlers.has(endpoint.path)) {
        throw new Error(
          `Multiple handlers registered for endpoint '${endpoint.path}'`
        );
      }
      this.handlers.set(endpoint.path, handler);
    };
  }
  async handle(path, request) {
    const handler = this.handlers.get(path);
    if (!handler) {
      return {
        kind: "error",
        message: `No handler is registered for path '${path}'`
      };
    }
    try {
      const response = await handler(request);
      return {
        kind: "success",
        response
      };
    } catch (e) {
      console.error(`Handler ${path} failed`, e);
      return {
        kind: "error",
        message: e.message
      };
    }
  }
}

const require$1 = createRequire(import.meta.url);
async function setupFrameworkPlugin({
  rootDirPath,
  frameworkPluginFactories
}) {
  const dependencies = await extractPackageDependencies(rootDirPath);
  for (const candidate of frameworkPluginFactories || []) {
    if (await candidate.isCompatible(dependencies)) {
      return candidate.create({
        rootDirPath,
        dependencies
      });
    }
  }
  return null;
}
async function extractPackageDependencies(rootDirPath) {
  const packageJsonPath = path.join(rootDirPath, "package.json");
  if (!await fs$1.pathExists(packageJsonPath)) {
    return {};
  }
  const { dependencies, devDependencies, peerDependencies } = JSON.parse(
    await fs$1.readFile(packageJsonPath, "utf8")
  );
  const allDependencies = {
    ...dependencies,
    ...devDependencies,
    ...peerDependencies
  };
  return Object.fromEntries(
    Object.entries(allDependencies).map(
      ([name, version]) => {
        let majorVersion;
        if (typeof version !== "string") {
          majorVersion = 0;
        } else if (version.startsWith("^") || version.startsWith("~")) {
          majorVersion = parseInt(version.slice(1));
        } else {
          majorVersion = parseInt(version);
        }
        const readInstalledVersion = async () => {
          try {
            const moduleEntryPath = findModuleEntryPath(name, rootDirPath);
            let packagePath = moduleEntryPath;
            let packageJsonPath2 = null;
            while (packagePath !== path.dirname(packagePath)) {
              const candidatePackageJsonPath = path.join(
                packagePath,
                "package.json"
              );
              if (fs$1.existsSync(candidatePackageJsonPath)) {
                packageJsonPath2 = candidatePackageJsonPath;
                break;
              }
              packagePath = path.dirname(packagePath);
            }
            if (!packageJsonPath2) {
              throw new Error(
                `No package.json path found from: ${moduleEntryPath}`
              );
            }
            const packageInfo = JSON.parse(
              await fs$1.readFile(packageJsonPath2, "utf8")
            );
            const version2 = packageInfo["version"];
            if (!version2 || typeof version2 !== "string") {
              throw new Error(
                `Invalid version found for package: ${packageJsonPath2}`
              );
            }
            return version2;
          } catch (e) {
            console.error(
              `Unable to read installed version of package: ${name}`,
              e
            );
            return null;
          }
        };
        return [name, { majorVersion, readInstalledVersion }];
      }
    )
  );
}
function findModuleEntryPath(name, rootDirPath) {
  try {
    return require$1.resolve(name, {
      paths: [rootDirPath]
    });
  } catch (e) {
    if (e.code === "ERR_PACKAGE_PATH_NOT_EXPORTED") {
      const potentialMatch = e.message.match(
        /No "exports" main defined in (.*)(\/|\\)package\.json/
      );
      if (potentialMatch) {
        return potentialMatch[1];
      }
    }
    throw e;
  }
}

const require = createRequire(import.meta.url);
process.on("uncaughtException", (e) => {
  console.error("Uncaught Exception:", e);
});
async function createWorkspace({
  rootDirPath,
  reader,
  frameworkPlugin,
  logLevel,
  setupEnvironment
}) {
  const expectedPluginApiVersion = 3;
  if (!frameworkPlugin.pluginApiVersion || frameworkPlugin.pluginApiVersion < expectedPluginApiVersion) {
    throw new Error(
      `Preview.js framework plugin ${frameworkPlugin.name} is incompatible with this version of Preview.js. Please upgrade it.`
    );
  } else if (frameworkPlugin.pluginApiVersion > expectedPluginApiVersion) {
    throw new Error(
      `Preview.js framework plugin ${frameworkPlugin.name} is too recent. Please upgrade Preview.js or use an older version of ${frameworkPlugin.name}.`
    );
  }
  if (frameworkPlugin.transformReader) {
    reader = frameworkPlugin.transformReader(reader);
  }
  const collected = {};
  const typeAnalyzer = createTypeAnalyzer({
    reader,
    rootDirPath,
    collected,
    specialTypes: frameworkPlugin.specialTypes,
    tsCompilerOptions: frameworkPlugin.tsCompilerOptions,
    printWarnings: logLevel === "info"
  });
  const router = new ApiRouter();
  router.registerRPC(RPCs.ComputeProps, async ({ componentIds }) => {
    let analyze;
    const detectedComponents = await frameworkPlugin.detectComponents(
      reader,
      typeAnalyzer,
      [
        ...new Set(
          componentIds.map(
            (c) => path.join(rootDirPath, decodeComponentId(c).filePath)
          )
        )
      ]
    );
    const componentIdToDetectedComponent = Object.fromEntries(
      detectedComponents.map((c) => [
        generateComponentId({
          filePath: path.relative(rootDirPath, c.absoluteFilePath),
          name: c.name
        }),
        c
      ])
    );
    const components = {};
    let types = {};
    for (const componentId of componentIds) {
      const component = componentIdToDetectedComponent[componentId];
      if (!component) {
        const { filePath, name } = decodeComponentId(componentId);
        throw new Error(`Component ${name} not detected in ${filePath}.`);
      }
      if (component.info.kind === "component") {
        analyze = component.info.analyze;
      } else {
        analyze = component.info.associatedComponent.analyze;
      }
      const { propsType: props, types: componentTypes } = await analyze();
      components[componentId] = {
        info: component.info.kind === "component" ? {
          kind: "component",
          exported: component.info.exported
        } : {
          kind: "story",
          args: component.info.args,
          associatedComponent: {
            filePath: path.relative(
              rootDirPath,
              component.info.associatedComponent.absoluteFilePath
            ),
            name: component.info.associatedComponent.name
          }
        },
        props
      };
      types = { ...types, ...componentTypes };
    }
    return {
      components,
      types
    };
  });
  router.registerRPC(
    RPCs.DetectComponents,
    (options) => detectComponents(workspace, frameworkPlugin, typeAnalyzer, options)
  );
  const middlewares = [
    express.json(),
    async (req, res, next) => {
      if (req.path.startsWith("/api/")) {
        res.json(await router.handle(req.path.substr(5), req.body));
      } else {
        next();
      }
    }
  ];
  const previewer = new Previewer({
    reader,
    rootDirPath,
    // TODO: Use a cleaner approach.
    previewDirPath: path.join(
      path.dirname(path.dirname(require.resolve("@previewjs/iframe"))),
      "preview"
    ),
    frameworkPlugin,
    logLevel,
    middlewares,
    onFileChanged: (absoluteFilePath) => {
      const filePath = path.relative(rootDirPath, absoluteFilePath);
      for (const name of Object.keys(collected)) {
        if (name.startsWith(`${filePath}:`)) {
          delete collected[name];
        }
      }
    }
  });
  const workspace = {
    rootDirPath,
    reader,
    typeAnalyzer,
    preview: {
      start: async (allocatePort) => {
        const port = await previewer.start(async () => {
          const port2 = allocatePort ? await allocatePort() : 0;
          return port2 || await getPort({
            port: portNumbers(3140, 4e3)
          });
        });
        return {
          url: () => `http://localhost:${port}`,
          stop: async (options) => {
            await previewer.stop(options);
          }
        };
      }
    },
    async localRpc(endpoint, request) {
      const result = await router.handle(endpoint.path, request);
      if (result.kind === "success") {
        return result.response;
      }
      throw new Error(result.message);
    },
    dispose: async () => {
      typeAnalyzer.dispose();
    }
  };
  if (setupEnvironment) {
    const environment = await setupEnvironment({
      registerRPC: (endpoint, handler) => router.registerRPC(endpoint, handler),
      workspace
    });
    if (environment.middlewares) {
      middlewares.push(...environment.middlewares);
    }
  }
  return workspace;
}
function findWorkspaceRoot(absoluteFilePath) {
  let dirPath = path.resolve(absoluteFilePath);
  while (dirPath !== path.dirname(dirPath)) {
    if (fs$1.existsSync(path.join(dirPath, "package.json")) && fs$1.existsSync(path.join(dirPath, "node_modules"))) {
      return dirPath;
    }
    dirPath = path.dirname(dirPath);
  }
  return null;
}

export { createWorkspace, findWorkspaceRoot, setupFrameworkPlugin };
